from __future__ import annotations

import asyncio
import asyncio.exceptions
import asyncio.timeouts
import contextlib
import datetime
import functools
import importlib
import logging
import pathlib
import sys
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, Iterator, Optional, cast

import aiormq.exceptions
from yarl import URL

from .app import Celery
from .context import CURRENT_ROOT_ID, CURRENT_TASK_ID
from .exceptions import MaxRetriesExceededError, Retry
from .request import Request
from .task import Task

if TYPE_CHECKING:
    import argparse

    from aio_pika import IncomingMessage

    from .annotated_task import AnnotatedTask

BANNER = """\
[config]
.> app:         {app}
.> transport:   {conninfo}
.> results:     {results}
.> concurrency: {concurrency}

[queues]
{queues}
"""

logger = logging.getLogger(__name__)


async def _sleep_if_necessary(task: Task) -> None:
    eta: datetime.datetime | None = task.request.eta
    if eta is None:
        return
    now = datetime.datetime.now().astimezone()
    if eta > now:
        delta: datetime.timedelta = eta - now
        logger.info(
            "[%s] Sleeping for %s until task ETA %s ...",
            task.request.id,
            delta,
            eta,
        )
        await asyncio.sleep(delta.total_seconds())


async def _handle_task_result(
    task: Task,
    annotated_task: AnnotatedTask,
    result: Any,
) -> None:
    if (
        task.app.conf.task_ignore_result
        or annotated_task.ignore_result
        or task.request.ignore_result
    ):
        return
    await task.update_state(state="SUCCESS", meta=result, _finalize=True)


async def _handle_task_retry(
    *,
    task: Task,
    annotated_task: AnnotatedTask,
    app: Celery,
    exc: Retry,
    message: IncomingMessage,
) -> None:
    logger.info(
        "Task %s[%s] retry: %s",
        task.name,
        task.request.id,
        exc,
    )
    max_retries = annotated_task.max_retries
    if max_retries is not None and task.request.retries > max_retries:
        msg = (
            f"Can't retry {task.name}[{task.request.id}] args:{task.request.args}"
            f" kwargs:{task.request.kwargs} max_retries:{max_retries}"
        )
        raise MaxRetriesExceededError(msg)
    await app.broker.publish_message(
        exc.message,
        routing_key=(message.routing_key or app.conf.task_default_queue),
    )


async def on_message_received(message: IncomingMessage, *, app: Celery) -> None:
    task_id: str | None = None
    task_name: str | None = None
    soft_time_limit = app.conf.task_soft_time_limit
    try:
        async with message.process(ignore_processed=True):
            task_id = str(message.headers["id"])
            task_name = str(message.headers["task"])
            root_id = cast(Optional[str], message.headers["root_id"])
            logger.info("Task %s[%s] received", task_name, task_id)
            CURRENT_ROOT_ID.set(root_id)
            CURRENT_TASK_ID.set(task_id)
            try:
                annotated_task = app.get_annotated_task(task_name)
            except KeyError:
                logger.exception(
                    "Received unregistered task of type %r\n"
                    "The message has been ignored and discarded.\n\n"
                    "Did you remember to import the module containing this task?\n"
                    "Or maybe you're using relative imports?\n\n"
                    "Please see\n"
                    "https://docs.celeryq.dev/en/latest/internals/protocol.html\n"
                    "for more information.\n\n"
                    "The full contents of the message body was:\n"
                    "%r (%db)\n\n"
                    "The full contents of the message headers:\n"
                    "%r\n\n"
                    "The delivery info for this task is:\n"
                    "%r",
                    task_name,
                    message.body,
                    len(message.body),
                    message.headers,
                    {
                        "consumer_tag": message.consumer_tag,
                        "delivery_tag": message.delivery_tag,
                        "redelivered": message.redelivered,
                        "exchange": message.exchange,
                        "routing_key": message.routing_key,
                    },
                )
                await message.reject()
                return

            task = Task(
                app=app,
                request=Request.from_message(message),
                _default_retry_delay=annotated_task.default_retry_delay,
            )
            await _sleep_if_necessary(task)

            timestamp_start = time.monotonic()
            args = (
                (task, *task.request.args) if annotated_task.bind else task.request.args
            )
            soft_time_limit = task.request.timelimit[0] or app.conf.task_soft_time_limit
            try:
                async with asyncio.timeouts.timeout(soft_time_limit):
                    result = await annotated_task.fn(*args, **task.request.kwargs)
            except Retry as exc:
                await _handle_task_retry(
                    task=task,
                    annotated_task=annotated_task,
                    app=app,
                    exc=exc,
                    message=message,
                )
            else:
                await _handle_task_result(
                    task=task,
                    annotated_task=annotated_task,
                    result=result,
                )
                next_message, next_routing_key = task.build_next_task_message(
                    result=result,
                )
                if next_message is not None:
                    await app.broker.publish_message(
                        next_message,
                        routing_key=next_routing_key,
                    )
                logger.info(
                    "Task %s[%s] succeeded in %.6fs: %r",
                    task.name,
                    task.request.id,
                    time.monotonic() - timestamp_start,
                    result,
                )
    except aiormq.exceptions.ChannelInvalidStateError:
        pass
    # This does not catch asyncio.exceptions.CancelledError since
    # the latter inherits directly from BaseException — not from Exception.
    # And that is okay for us. CancelledError is not meant to be caught.
    except Exception as err:
        if isinstance(err, asyncio.exceptions.TimeoutError):
            logger.warning(
                "Soft time limit (%ss) exceeded for %s[%s]",
                soft_time_limit,
                task_name,
                task_id,
            )
        logger.exception(
            "Task %s[%s] raised unexpected:",
            task_name,
            task_id,
        )
    finally:
        CURRENT_ROOT_ID.set(None)
        CURRENT_TASK_ID.set(None)


@contextlib.contextmanager
def _cwd_in_path() -> Iterator[None]:
    """Context adding the current working directory to sys.path."""
    cwd = str(pathlib.Path.cwd())
    if cwd in sys.path:
        yield
    else:
        sys.path.insert(0, cwd)
        try:
            yield
        finally:
            with contextlib.suppress(ValueError):
                sys.path.remove(cwd)


def _find_app_instance(locator: str) -> Celery:
    with _cwd_in_path():
        module_name, app_name = locator.rsplit(":", 1)
        module = importlib.import_module(module_name)
        app = getattr(module, app_name)
    if not isinstance(app, Celery):
        msg = "app instance must inherit from Celery class"
        raise TypeError(msg)
    return app


def _print_intro(
    concurrency: int,
    queue_names: list[str],
    task_names: list[str],
    app: Celery,
) -> None:
    q = ".> " + "\n.> ".join(queue_names)
    t = "  . " + "\n  . ".join(task_names)

    def _normalize(u: str | None) -> str:
        if not u:
            return "disabled://"
        parts = urllib.parse.urlparse(u)
        if not parts.password:
            return u
        return str(URL(u).with_password("******"))

    banner = BANNER.format(
        app=repr(app),
        conninfo=_normalize(app.conf.broker_url),
        results=_normalize(app.conf.result_backend),
        concurrency=f"{concurrency} (asyncio)",
        queues=q,
    ).splitlines()
    print(  # noqa: T201
        "╔═══════════════════════╗\n"
        "║ AsyncIO Celery Worker ║\n"
        "╚═══════════════════════╝\n",
    )
    print("\n".join(banner) + "\n")  # noqa: T201
    print(f"[tasks]\n{t}\n")  # noqa: T201


async def run(args: argparse.Namespace) -> None:
    logging.basicConfig(
        level=logging.getLevelName(args.loglevel),
        format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
    )
    logging.getLogger("aio_celery").setLevel(level=logging.getLevelName(args.loglevel))
    app = _find_app_instance(args.app)
    if not args.queues:
        queue_names = [app.conf.task_default_queue]
    else:
        queue_names = [q.strip() for q in args.queues.split(",")]

    _print_intro(
        args.concurrency,
        queue_names,
        app.list_registered_task_names(),
        app,
    )

    async with app.setup():
        await app.broker.set_qos(prefetch_count=args.concurrency)
        queues = [await app.broker.declare_queue(name) for name in queue_names]
        for queue in queues:
            await queue.consume(functools.partial(on_message_received, app=app))
        logger.info("Waiting for messages. To exit press CTRL+C")
        await asyncio.Future()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime

    from aio_pika import Message


class CeleryError(Exception):
    """Base class for all Celery errors."""


class MaxRetriesExceededError(CeleryError):
    """The tasks max restart limit has been exceeded."""


class TimeoutError(CeleryError):  # noqa: A001
    """The operation timed out."""


class Retry(CeleryError):  # noqa: N818
    def __init__(self, *, message: Message, delay: datetime.timedelta) -> None:
        self.message = message
        self._delay = delay
        super().__init__()

    def __str__(self) -> str:
        return f"{self.__class__.__name__} in {self._delay.total_seconds()}s"



from __future__ import annotations

import dataclasses
import datetime
from dataclasses import dataclass


@dataclass
class DefaultConfig:
    # Task execution settings
    task_ignore_result: bool = False
    task_soft_time_limit: int | None = None

    # Task result backend settings
    result_backend: str | None = None
    result_backend_connection_pool_size: int = 50
    result_expires: datetime.timedelta = datetime.timedelta(days=1)

    # Message Routing
    task_default_priority: int | None = None
    task_default_queue: str = "celery"
    task_queue_max_priority: int | None = None

    # Broker Settings
    broker_url: str = "amqp://guest:guest@localhost:5672//"

    def update(self, **options: int | bool | str | datetime.timedelta | None) -> None:
        fields = {f.name for f in dataclasses.fields(self.__class__)}
        for k, v in options.items():
            if k not in fields:
                msg = f"Unknown configuration option: {k!r}"
                raise ValueError(msg)
            setattr(self, k, v)



from __future__ import annotations

import contextlib
import logging
import sys
import uuid
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
)

import aio_pika

from .amqp import create_task_message
from .annotated_task import AnnotatedTask
from .backend import create_redis_connection_pool
from .broker import Broker
from .config import DefaultConfig
from .context import CURRENT_ROOT_ID, CURRENT_TASK_ID
from .result import AsyncResult as _AsyncResult
from .utils import first_not_null

if TYPE_CHECKING:
    import redis.asyncio

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _CompleteTaskResources:
    context: Any
    redis_client_celery: Any


class Celery:
    def __init__(self) -> None:
        self.conf = DefaultConfig()
        self._tasks_registry: dict[str, AnnotatedTask] = {}
        self._app_context: Any = None
        self._result_backend_connection_pool: (
            redis.asyncio.BlockingConnectionPool | None
        ) = None
        self._broker: Broker | None = None
        self._setup_app_context: Callable[
            [],
            contextlib.AbstractAsyncContextManager[Any],
        ] = _setup_nothing

    def define_app_context(
        self,
        fn: Callable[[], contextlib.AbstractAsyncContextManager[Any]],
    ) -> None:
        self._setup_app_context = fn

    @property
    def broker(self) -> Broker:
        if self._broker is None:
            msg = "This app has not been configured yet."
            raise RuntimeError(msg)
        return self._broker

    @broker.setter
    def broker(self, b: Broker) -> None:
        self._broker = b

    @property
    def context(self) -> Any:
        return self._app_context

    @property
    def result_backend(self) -> redis.asyncio.Redis[bytes] | None:
        if self._result_backend_connection_pool is None:
            return None

        from redis.asyncio import Redis

        return Redis(connection_pool=self._result_backend_connection_pool)

    @contextlib.asynccontextmanager
    async def setup(self) -> AsyncIterator[None]:
        connection = await aio_pika.connect_robust(self.conf.broker_url)
        async with connection, connection.channel() as channel:
            self.broker = Broker(
                rabbitmq_channel=channel,
                task_queue_max_priority=self.conf.task_queue_max_priority,
            )
            try:
                if self.conf.result_backend is not None:
                    self._result_backend_connection_pool = create_redis_connection_pool(
                        url=self.conf.result_backend,
                        pool_size=self.conf.result_backend_connection_pool_size,
                    )
                async with self._setup_app_context() as context:
                    self._app_context = context
                    try:
                        yield
                    finally:
                        logger.warning("Shutting down application.")
            finally:
                if self._result_backend_connection_pool is not None:
                    await self._result_backend_connection_pool.disconnect()

    def task(  # noqa: PLR0913
        self,
        *args: Callable[..., Any],
        bind: bool = False,
        name: str | None = None,
        ignore_result: bool | None = None,
        max_retries: int | None = 3,
        default_retry_delay: int = 180,
        queue: str | None = None,
        priority: int | None = None,
    ) -> AnnotatedTask | Callable[[Callable[..., Any]], AnnotatedTask]:
        """Create a task class out of any callable."""

        def decorator(fn: Callable[..., Any]) -> AnnotatedTask:
            if name is None:
                task_name = _gen_task_name(fn.__name__, fn.__module__)
            else:
                task_name = name
            annotated_task = AnnotatedTask(
                fn=fn,
                bind=bind,
                ignore_result=ignore_result,
                max_retries=max_retries,
                default_retry_delay=default_retry_delay,
                name=task_name,
                queue=queue,
                priority=priority,
                app=self,
            )
            self._tasks_registry[task_name] = annotated_task
            return annotated_task

        if not args:
            return decorator
        if len(args) == 1:
            func = args[0]
            if not callable(func):
                msg = "argument 1 to @task() must be a callable"
                raise TypeError(msg)
            return decorator(func)
        msg = "@task() takes exactly 1 argument"
        raise TypeError(msg)

    def get_annotated_task(self, task_name: str) -> AnnotatedTask:
        return self._tasks_registry[task_name]

    def list_registered_task_names(self) -> list[str]:
        return sorted(self._tasks_registry)

    def AsyncResult(self, task_id: str) -> _AsyncResult:  # noqa: N802
        return _AsyncResult(task_id, app=self)

    async def send_task(  # noqa: PLR0913
        self,
        name: str,
        *,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        countdown: float | None = None,
        task_id: str | None = None,
        priority: int | None = None,
        queue: str | None = None,
    ) -> _AsyncResult:
        if task_id is None:
            task_id = str(uuid.uuid4())
        await self.broker.publish_message(
            create_task_message(
                task_id=task_id,
                task_name=name,
                args=args,
                kwargs=kwargs,
                priority=priority,
                countdown=countdown,
                parent_id=CURRENT_TASK_ID.get(),
                root_id=CURRENT_ROOT_ID.get(),
            ),
            routing_key=first_not_null(queue, self.conf.task_default_queue),
        )
        return self.AsyncResult(task_id)


@contextlib.asynccontextmanager
async def _setup_nothing() -> AsyncIterator[None]:
    yield None


def _gen_task_name(name: str, module_name: str) -> str:
    """Generate task name from name/module pair."""
    module_name = module_name or "__main__"
    module = sys.modules[module_name]
    if module is not None:
        module_name = module.__name__
    return ".".join(p for p in (module_name, name) if p)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    import redis.asyncio


def create_redis_connection_pool(
    *,
    url: str,
    pool_size: int,
) -> redis.asyncio.BlockingConnectionPool:
    import redis.backoff
    from redis.asyncio import BlockingConnectionPool
    from redis.asyncio.retry import Retry

    default_base = getattr(redis.backoff, "DEFAULT_BASE", 0.008)
    default_cap = getattr(redis.backoff, "DEFAULT_CAP", 0.512)

    return cast(
        BlockingConnectionPool,
        BlockingConnectionPool.from_url(
            url=url,
            max_connections=pool_size,
            timeout=None,
            socket_timeout=5,
            retry_on_timeout=True,
            retry=Retry(
                backoff=redis.backoff.EqualJitterBackoff(
                    cap=default_cap,
                    base=default_base,
                ),
                retries=5,
            ),
        ),
    )



from __future__ import annotations

import copy
import datetime
import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, cast

if TYPE_CHECKING:
    from aio_pika import IncomingMessage, Message


@dataclass(frozen=True, kw_only=True)
class Request:
    message: IncomingMessage

    id: str  # noqa: A003
    task: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    retries: int
    eta: datetime.datetime | None
    parent_id: str | None
    group: str | None
    root_id: str | None
    ignore_result: bool
    timelimit: tuple[int | None, int | None]
    chain: list[dict[str, Any]]

    @classmethod
    def from_message(cls: type[Request], message: IncomingMessage) -> Request:
        headers = message.headers
        args, kwargs, options = json.loads(message.body)
        eta = headers["eta"]
        if eta is not None:
            eta = datetime.datetime.fromisoformat(str(eta))
        return cls(
            message=message,
            id=str(headers["id"]),
            args=args,
            kwargs=kwargs,
            task=cast(str, headers["task"]),
            retries=cast(int, headers["retries"]),
            eta=eta,
            parent_id=cast(Optional[str], headers["parent_id"]),
            group=cast(Optional[str], headers["group"]),
            root_id=cast(str, headers["root_id"]),
            ignore_result=bool(headers["ignore_result"]),
            timelimit=cast(tuple[Optional[int], Optional[int]], headers["timelimit"]),
            chain=options["chain"],
        )

    def build_retry_message(
        self,
        *,
        countdown: datetime.timedelta | None,
    ) -> Message:
        message = copy.copy(self.message)
        if countdown is None:
            eta = None
        else:
            eta = (datetime.datetime.now().astimezone() + countdown).isoformat()
        message.headers.update(
            eta=eta,
            retries=self.retries + 1,
            parent_id=self.id,
        )
        return message



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aio_pika import Message
    from aio_pika.abc import AbstractChannel, AbstractQueue
    from pamqp.common import Arguments


class Broker:
    def __init__(
        self,
        *,
        rabbitmq_channel: AbstractChannel,
        task_queue_max_priority: int | None,
    ) -> None:
        self._rabbitmq_channel = rabbitmq_channel
        self._task_queue_max_priority: int | None = task_queue_max_priority
        self._already_declared_queues: set[str] = set()

    async def publish_message(
        self,
        message: Message,
        *,
        routing_key: str,
    ) -> None:
        if routing_key not in self._already_declared_queues:
            await self.declare_queue(routing_key)
            self._already_declared_queues.add(routing_key)
        await self._rabbitmq_channel.default_exchange.publish(
            message,
            routing_key=routing_key,
            timeout=60,
        )

    async def declare_queue(self, queue_name: str) -> AbstractQueue:
        arguments: Arguments
        if self._task_queue_max_priority is not None:
            arguments = {"x-max-priority": self._task_queue_max_priority}
        else:
            arguments = None
        queue = await self._rabbitmq_channel.declare_queue(
            name=queue_name,
            durable=True,
            arguments=arguments,
        )
        self._already_declared_queues.add(queue_name)
        return queue

    async def set_qos(self, *, prefetch_count: int) -> None:
        await self._rabbitmq_channel.set_qos(prefetch_count=prefetch_count)
